/*-- astronomical calculation library --
1991/12/31 T.Handa (create)
2007/11/14 T.Handa (add) calc_lst2()
2007/11/20 T.Handa (add) B1950toJ2000(),J2000toB1950(),B1950toJ2000_dircos(),J2000toB1950_dircos()
2007/12/05 T.Handa (add) lb2radec50(),radec502lb()
*/
/* celestial calculation */
/*=== macro definitions ===*/
#define SIZE_SITENAME 9

/*=== structures ===*/
typedef struct {
  double d; /* ephemeris time from 1899/12/31 12:00(ET) [day] */
  double t; /* ephemeris time from 1899/12/31 12:00(ET) [Julian century] */
  double t50; /* ephemeris time from B1950.0 [Besselian century] */
} EP_TIME;

#ifndef XY_POS_DEFINE
typedef struct {
  double x; 	/* longitude [degree] */
  double y; 	/* latitude [degree] */
} XY_POS;
#define XY_POS_DEFINE
#endif

typedef struct {
  double x_rad; /* longitude [rad] */
  double y_rad; /* latitude [rad] */
} XY_POS_RAD;

typedef struct {
  double l; 	/* x-component of direct cosine */
  double m; 	/* y-component of direct cosine */
  double n; 	/* z-component of direct cosine */
} DIRECT_COS;

typedef struct {
  double d_psi; /* nutation of ecliptic longitude psi[rad] */
  double d_epsilon; /* nutation of inclination of the ecliptic plane epsilon[rad] */
  double epsilon; /* averaged inclination of the ecliptic plane epsilon[rad] */
} NUTATION_PRM;

typedef struct {
  char name[SIZE_SITENAME]; 	/* name of the site */
  double x; 	/* longitude of site [degree]: eastbound is positive */
  double y; 	/* latitude of site [degree]: northword is positive */
  double z; 	/* height above sea level[m] */
  double gravity; 	/* gravity at the site [Gal=cm/s/s] */
} SITE_POS;

typedef struct {
  double tmp; /* temperature [degree C] */
  double water; /* partial pressure of water vaper [hPa] */
  double press; /* atmospheric pressure at 1.5m above the ground [hPa] */
} METEO_PRM;

#ifndef time_t
#define time_t long
#endif

/*=== function prototype ===*/
#ifdef _NO_PROTO
extern void apparent();
extern void aberration();
extern void nutation();
extern void precession();
extern void ab_param();
extern void nut_param();

extern time_t get_leapsec();
extern int ephem_time();
extern double besselian();
extern double mean_sun();
extern long m_julian_day();

extern void dir_cos();
extern void direction();

extern double calc_lst();
extern double calc_lst2();
extern double calc_lst_hour();

extern void radec2azel();
extern double refract();
extern double refract_optical();
extern double gravity1980();

extern void trans_sphere();
extern int lb2radec50();
extern int radec502lb();
extern int B1950toJ2000_dircos();
extern int B1950toJ2000();
extern int J2000toB1950_dircos();
extern int J2000toB1950();

extern double angle_distance();

#else
extern void apparent(EP_TIME, DIRECT_COS, DIRECT_COS *, NUTATION_PRM *);
extern void aberration(DIRECT_COS, double, double, DIRECT_COS *);
extern void nutation(DIRECT_COS, NUTATION_PRM, DIRECT_COS *);
extern void precession(DIRECT_COS, double, DIRECT_COS *);
extern void ab_param(double, double *);
extern void nut_param(EP_TIME, NUTATION_PRM *);

extern time_t get_leapsec(time_t);
extern int ephem_time(long, EP_TIME *);
extern double besselian(long, double);
extern double mean_sun(double);
extern long m_julian_day(long);

extern void dir_cos(XY_POS, DIRECT_COS *);
extern void direction(DIRECT_COS, XY_POS *);

extern double calc_lst(long, SITE_POS, NUTATION_PRM);
extern double calc_lst2(double,SITE_POS,NUTATION_PRM);
extern double calc_lst_hour(long, SITE_POS);

extern void radec2azel(double, DIRECT_COS, SITE_POS, XY_POS *);
extern double refract(double, METEO_PRM);
extern double refract_optical(double, double, METEO_PRM, double);
extern double gravity1980(double);

extern void trans_sphere(double, double, double, XY_POS_RAD, XY_POS_RAD *);
extern int lb2radec50(XY_POS *,XY_POS);
extern int radec502lb(XY_POS *,XY_POS);
extern int B1950toJ2000_dircos(DIRECT_COS *,DIRECT_COS);
extern int B1950toJ2000(XY_POS *,XY_POS);
extern int J2000toB1950_dircos(DIRECT_COS *,DIRECT_COS);
extern int J2000toB1950(XY_POS *,XY_POS);

extern double angle_distance(XY_POS, XY_POS);
#endif

/* velocity calculation */
#ifdef _NO_PROTO
extern double vel_shift();
extern double v_solar_motion();
extern double v_earth_spin();
extern double v_earth_revolution();
extern double v_earth_moon();

#else 
extern double vel_shift(EP_TIME, double, XY_POS, int, DIRECT_COS, SITE_POS, double *);
extern double v_solar_motion(EP_TIME, int, DIRECT_COS);
extern double v_earth_spin(double, XY_POS, SITE_POS);
extern double v_earth_revolution(EP_TIME, XY_POS, XY_POS_RAD *);
extern double v_earth_moon(EP_TIME, XY_POS_RAD);
#endif

/* spherical transform */
#ifdef _NO_PROTO
extern double dircos_inner_product();
extern void dircos_outer_product();
extern double VectorAngleOnUnitSphere();
extern double DistanceOnUnitSphere();
extern double AzEl2PA();

#else /* undef _NO_PROTO */
extern double dircos_inner_product(DIRECT_COS, DIRECT_COS);
extern void dircos_outer_product(DIRECT_COS *, DIRECT_COS, DIRECT_COS);
extern double VectorAngleOnUnitSphere(DIRECT_COS, DIRECT_COS, DIRECT_COS);
extern double DistanceOnUnitSphere(DIRECT_COS, DIRECT_COS);
extern double AzEl2PA(XY_POS, XY_POS);
#endif /* endif of define _NO_PROTO */

/* for astrometric caliculation */
#ifdef _NO_PROTO
extern void xy2azel(XY_POS *, EP_TIME, double, SITE_POS, int, XY_POS);

#else 
extern void xy2azel();
#endif
