/*
STAR Astronomical calculation functions

1997/11/13 T.Handa
1999/5/9 T.Handa (revise) off_line process for leap_sec.
TO_DO Y2K
2000/9/28 T.Handa (revise) date to resolve the Y2K problem
2005/4/17 T.Handa (debug) argument order of aberration() in apparent()
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include <StData.h>
#include <StStd.h>

#define SIZE_DM 2048

/*
astronomical calculation library; beside of ASTROS

1995/7/24 T.Handa
1996/5/2 T.Handa add function
	dircos_inner_product,dircos_outer_product,VectorAngleOnUnitSphere
1996/5/9 T.Handa add function
	DistanceOnUnitSphere
1996/5/22 T.Handa add function
	stod
*/

int month_name2number(month_name)
char *month_name;
/*
convert month name to its number
check only first 3 letters. The first letter must be uppercase

2001/9/26 T.Handa
*/
{
  static char month_str[12][4]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
  int ii;

  for(ii=0;ii<12;ii++) {
    if(strncmp(month_name,month_str[ii],3)==0) return(ii+1);
  }

  return(0);
}

int isleapyear(year)
int year;
/*
check the leap year in Gregorian rule

2001/9/26 T.Handa
*/
{
  int leap_year_flag;
  int y4,y100,y400;

  /* check the leap year in Gregorian rule */
  y4=year/4;
  y100=year/100;
  y400=year/400;

  if(y400*400==year) {
    leap_year_flag=1;
  } else if(y100*100==year) {
    leap_year_flag=0;
  } else if(y4*4==year) {
    leap_year_flag=1;
  } else {
    leap_year_flag=0;
  }

  return(leap_year_flag);
}

int dayfromJan00(year,month,day)
int year,month,day;
/*
convert year month day to the date from Jan 0
with leap year condition
output for 2001/3/21 is 60, output for 2000/3/21 is 61
output for 2001/1/1 is 1
also valid for nominal date out of month length; e.g. 2001/1/0 and 2001/3/32

2001/9/26 T.Handa
*/
{
  int leap_year_flag;

  static int day_of_month[12]={31,28,31,30,31,30,31,31,30,31,30,31};
  int day_from_top;
  int days;
  int ii;

  /* check the leap year in Gregorian rule */
  leap_year_flag=isleapyear(year);

  /* month of the day */
  days=0;

  for(ii=0;ii<month-1;ii++) {
    days=days+day_of_month[ii];
  }

  /* adjust the leap year */
  if(month>2) days=days+leap_year_flag;

  day_from_top=days+day;

  return(day_from_top);
}

int fromJan002date(year,days,month,day)
int year,days;
int *month,*day;
/*
convert from month day to the date from Jan 0 in the year
with leap year condition
output for 60 in 2001 is 3/21, output for 60 in 2000 is 3/20
output for 1 in 2001 is 1/1

2001/9/27 T.Handa
*/
{
  int leap_year_flag;

  static int day_of_month[12]={31,28,31,30,31,30,31,31,30,31,30,31};
  static int day_in_feb[2]={28,29};
  int ii;

  /* check the leap year in Gregorian rule */
  leap_year_flag=isleapyear(year);

  *month=0;
  *day=0;

  if(leap_year_flag==1) {
    if(days>366) return 100;
  } else {
    if(days>365) return 100;
  }

  /* adjust the leap year */
  day_of_month[1]=day_in_feb[leap_year_flag];

  /* month of the day */
  for(ii=0;ii<12;ii++) {
    if(days<=day_of_month[ii]) {
      *month=ii+1;
      break;
    }
    days=days-day_of_month[ii];
  }

  *day=days;

  return(0);
}

long ymd2mjd(year,month,day)
/*
convert from date in ephemerical format to modified Julian day 
MJD of 1999/12/31 is jday=51543
valid after 1901/1/1

2001/9/26 T.Handa
*/
int year,month,day;
{
  long fromnyday,mjd;
  int ii;

  fromnyday=(long)dayfromJan00(year,month,day);

  mjd=15020L;

  for(ii=1900;ii<year;ii++) {
    if(isleapyear(ii)==1) mjd=mjd+366L;
    else mjd=mjd+365L;
  }

  mjd=mjd+fromnyday;

  return(mjd);
}

int mjd2ymd(mjd,year,month,day)
/*
convert from modified Julian day to date in ephemerical format 
MJD=51543 is 1999/12/31
valid after 1901/1/1

2001/9/27 T.Handa
*/
long mjd;
int *year,*month,*day;
{
  long fromnyday;
  int yy,ret_code;
  static int daysinayear[2]={365,366};

  fromnyday=mjd-15020L;
  yy=1900;

  while(fromnyday>daysinayear[isleapyear(yy)]) {
    fromnyday=fromnyday-daysinayear[isleapyear(yy)];
    yy++;
  }
  *year=yy;

  ret_code=fromJan002date(yy,fromnyday,month,day);

  return ret_code;
}

long mjd2ephem(jday)
long jday;
/*
convert from modified Julian day to date in ephemerical format 
correct from iday=000401,jday=15110 to iday=991231,jday=51543

1985/5/5 T.Handa : create
1988/2/12 T.Handa : revise
1995/7/24 T.Handa : convert from FORTRAN to C
*/
{
	long	iday;
	long	jd,jx,jm;
	long	iyy,imm,idd;

	jd=jday-15078;
	iyy=(4*jd-1)/1461;
	jx=iyy*1461/4;
	jx=jd-jx;
	imm=5*jx-3;
	imm=imm/153;
	jm=imm/10;
	iyy=iyy+jm;
	idd=5*jx+2-153*imm;
	idd=idd/5;
	imm=imm+jm*12+3;
	imm=imm-(long)((imm-1)/12)*12;
	iday=iyy*10000+imm*100+idd;
	return(iday);
}


long ephem2mjd(iday)
long iday;
/*
convert from date in ephemerical format to modified Julian day 
correct from iday=000401,jday=15110 to iday=991231,jday=51543

1985/5/5 T.Handa : create
1988/2/12 T.Handa : revise
1995/7/24 T.Handa : convert from FORTRAN to C
*/
{
	long	jday;
	long	jd,jx,jm;
	long	iyy,imm,idd;

	iyy=iday/10000;
	idd=iday-iyy*10000;
	imm=idd/100;
	idd=idd-imm*100;
	jx=(14-imm)/12;
	jm=jx*12+imm-2;
	jm=jm*367;
	jm=jm/12;
	jd=(-jx)+6700+iyy;
	jx=jd/100+1;
	jx=jx*3;
	jx=jx/4;
	jd=jd*1461;
	jd=jd/4;
	jd=jd-2432076;
	jday=jd+jm-jx+idd;

	return(jday);
}

long mjd2time1970(mjday,timesec)
long mjday;	/* Modified Julian Day */
long timesec;	/* time in sec from 0:00 */
/*
Calcularte UNIX sequencial time from 1970 Jan. 1 0:00
from MJD and sequencial time
1970/1/1 0:00 UTC = 40587.0 MJD

1995/7/24 T.Handa
*/
{
  long time1970;

  time1970=(mjday-40587)*86400+timesec;
  return(time1970);
}

long month_number(month_name)
char month_name[];
/*
Convert from month name to serial number of the month
say Jan --> 1, Feb --> 2

1996/1/23 T.Handa
1997/11/27 T.Handa (debug for Dec-->12)

should be use int month_name2number();
*/
{
  long month;

  if(strncmp(month_name,"Jan",3)==0)
    month=1;
  else if(strncmp(month_name,"Feb",3)==0)
    month=2;
  else if(strncmp(month_name,"Mar",3)==0)
    month=3;
  else if(strncmp(month_name,"Apr",3)==0)
    month=4;
  else if(strncmp(month_name,"May",3)==0)
    month=5;
  else if(strncmp(month_name,"Jun",3)==0)
    month=6;
  else if(strncmp(month_name,"Jul",3)==0)
    month=7;
  else if(strncmp(month_name,"Aug",3)==0)
    month=8;
  else if(strncmp(month_name,"Sep",3)==0)
    month=9;
  else if(strncmp(month_name,"Oct",3)==0)
    month=10;
  else if(strncmp(month_name,"Nov",3)==0)
    month=11;
  else if(strncmp(month_name,"Dec",3)==0)
    month=12;
  else
    month=0;
  return(month);
}

long ctime2utc(ctime_char)
/*
convert time format characters (ctime output) into UTC from 1 Jan 1970 0:0:0

1996/1/23 T.Handa
*/
char ctime_char[]; /* <in> format chars "Tue Jan 23 20:22:42 1996\n" */
{
  long utc;
  DATE_AND_TIME	date;

  ctime2date_and_time(ctime_char, &date);
  utc=date_and_time2utc(date);
  return(utc);
}

void ctime2date_and_time(ctime_char, date)
/*
convert time format characters (ctime output) into DATE_AND_TIME format

1996/1/23 T.Handa
2001/9/28 T.Handa (month_number() --> month_name2number() )
*/
char ctime_char[];	/* <in> format chars "Tue Jan 23 20:22:42 1996\n" */
DATE_AND_TIME *date;	/* <out> time structure */
{
  char buffer[25],month_name[4];

  strncpy(buffer,ctime_char,24);
  buffer[24]=(char)NULL;

  strncpy(month_name,&(buffer[4]),3);
  month_name[3]=(char)NULL;
  (*date).month=(long)month_name2number(month_name);
  buffer[10]=(char)NULL;
  (*date).day=atol(&(buffer[8]));
  buffer[13]=(char)NULL;
  (*date).hour=atol(&(buffer[11]));
  buffer[16]=(char)NULL;
  (*date).min=atol(&(buffer[14]));
  buffer[19]=(char)NULL;
  (*date).sec=atol(&(buffer[17]));
  (*date).year=atol(&(buffer[20]));

  return;
}

long date_and_time2utc(target_date)
DATE_AND_TIME target_date;
/*
convert date in DATE_AND_TIME format to UTC from 1 Jan 1970 0:0:0

1996/1/23 T.Handa
*/
{
  long utc;
  DATE_AND_TIME trial_date;

/* initial candidate */
  utc=((target_date.year-1970)*365
       +(target_date.month-1)*30+target_date.day)*24L*3600;
  utc=utc+((target_date.hour*24L+target_date.min)*60L+target_date.sec);

  while(1){
    ctime2date_and_time(ctime(&utc), &trial_date);

    if(target_date.year<trial_date.year)
      utc=utc-365L*24L*3600L;
    else if(target_date.year>trial_date.year)
      utc=utc+365L*24L*3600L;
    else {
      if(target_date.month<trial_date.month)
	utc=utc-30L*24L*3600L;
      else if(target_date.month>trial_date.month)
	utc=utc+30L*24L*3600L;
      else {
	if(target_date.day<trial_date.day)
	  utc=utc-24L*3600L;
	else if(target_date.day>trial_date.day)
	  utc=utc+24L*3600L;
	else {
	  if(target_date.hour<trial_date.hour)
	    utc=utc-3600L;
	  else if(target_date.hour>trial_date.hour)
	    utc=utc+3600L;
	  else {
	    if(target_date.min<trial_date.min)
	      utc=utc-60L;
	    else if(target_date.min>trial_date.min)
	      utc=utc+60L;
	    else {
	      if(target_date.sec<trial_date.sec)
		utc=utc-1L;
	      else if(target_date.sec>trial_date.sec)
		utc=utc+1L;
	      else
		return(utc);
	    }}}}}
  }
}
