/*
STAR Profile data estimating functions

1997/12/25 T.Handa
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>

#include <StData.h>
#include <StStd.h>

/* -------PROFILE data estimation -------*/

double StarCalcRms(profile,ch_range)
/*
calculate rms noise level

return:expected rms for mother distribution
profile.rx_status.rmssimple rms

1992/12/22 Naomi Hirano
1997/12/25 T.Handa (ver.1.10) completely revised, simplify+input profile.rx_status
1998/1/5 T.Handa (for rx_status.scale)
*/
PROFILE *profile;	/* <i/o> */
RANGE_long ch_range;	/* channel range for rms calculation, must be sorted */
{
  int ch,n_ch;
  double rms,mean_level,deviate;
  int spurious_flag,baseline_flag;
  double scale;

/* parameter initialize */
  n_ch=0;
  mean_level=0.0;
  rms=0.0;
  scale=(*profile).rx_status.scale;

/* estimate the average level */
  for(ch=ch_range.min;ch<=ch_range.max;ch++) {
    spurious_flag=StarReadSpuriousFlag(profile,ch);
    baseline_flag=StarReadBaselineFlag(profile,ch);
    if(spurious_flag==FLAGGED)
      StarWriteBaselineFlag(profile,ch,UNFLAGGED);
    else if(baseline_flag==FLAGGED) {
      n_ch++;
      mean_level=mean_level+(double)((*profile).data[ch-1]);
    }
  }

  if(n_ch==0) return(0.0);

  mean_level=mean_level/(double)n_ch;

/* estimate the root mean square of deviations */
  for(ch=ch_range.min;ch<=ch_range.max;ch++) {
    baseline_flag=StarReadBaselineFlag(profile,ch);
    if(baseline_flag==FLAGGED) {
      deviate=(double)((*profile).data[ch-1])-mean_level;
      rms=rms+deviate*deviate;
    }
  }
  (*profile).rx_status.rms=(float)(sqrt(rms/(double)n_ch))*scale;

  if(n_ch<=1) return(0.0);
  mean_level=mean_level*scale;
  rms=sqrt(rms/(double)(n_ch-1))*scale;
  return(rms);
}

void StarProfileMaxMin(profile,max_ch,min_ch,max,min)
/*
estimate max and min of profile data

1998/1/5 T.Handa (for rx_status.scale)
*/
PROFILE *profile;	/* <in>  */
int max_ch,min_ch;	/* <in>  */
double *max, *min;	/* <out> */
{
  int ii;
  double y_max,y_min;
  double data;
  double scale;

  if(max_ch<min_ch) {
    ii=max_ch;
    max_ch=min_ch;
    min_ch=ii;
  }

  scale=(*profile).rx_status.scale;

  y_max=(*profile).data[min_ch-1];
  y_min=(*profile).data[min_ch-1];
  for(ii=min_ch;ii<=max_ch;ii++) {
    if(StarReadSpuriousFlag(profile,ii)==UNFLAGGED) {
      data=(*profile).data[ii-1];
      if(data<y_min) y_min=data;
      if(data>y_max) y_max=data;
    }
  }
  *max=y_max*scale; *min=y_min*scale;
}

int StarEstimateActualChRange(profile,x_range,x_unit,act_ch_range)
PROFILE *profile; /* <in> */
RANGE x_range; /* <in> */
int x_unit; /* <in> */
RANGE_long *act_ch_range; /* <out> */
/*
estimate channel range correponding x_range in x_unit
act_ch_range is sorted, which means act_ch_range.min is not always
correspond to x_range.min

1997/12/10 T.Handa
1999/ 8/ 6 T.Handa (debug) act_ch_range for OVER_FLOW/UNDER_FLOW
*/
{
  RANGE_long ch_range;
  int error_code;

/* set range along the x-axis in the proper physical unit */
  if(x_unit==X_CHANNEL) {
    ch_range.min=(long)x_range.min; ch_range.max=(long)x_range.max;
    if(ch_range.min==ch_range.max) {
      ch_range.min=1; ch_range.max=(*profile).be_status.data_ch;
    }
  } else if(x_unit==X_RF_FREQ) {
    ch_range.min=(long)StarFrequency2Ch(profile,&error_code,x_range.min);
    if(error_code!=NORMAL_END) {
      //      fprintf(stderr,"profile=%s:",(*profile).file_name);
      switch(error_code) {
      case UN_CONVERGE:
	fprintf(stderr,"profile=%s:",(*profile).file_name);
	fprintf(stderr,"Unconverge");
	fprintf(stderr," at RF=%f\n",x_range.min);
	ch_range.min=1;
	break;
      case UNDER_FLOW:
	//fprintf(stderr,"UnderFlow");
	ch_range.min=1;
	break;
      case OVER_FLOW:
	//fprintf(stderr,"OverFlow");
	ch_range.min=(*profile).be_status.data_ch;
	break;
      }
      //fprintf(stderr," at RF=%f\n",x_range.min);
    }
    ch_range.max=(long)StarFrequency2Ch(profile,&error_code,ceil(x_range.max));
    if(error_code!=NORMAL_END) {
      //      fprintf(stderr,"profile=%s:",(*profile).file_name);
      switch(error_code) {
      case UN_CONVERGE:
	fprintf(stderr,"profile=%s:",(*profile).file_name);
	fprintf(stderr,"Unconverge");
	fprintf(stderr," at RF=%f\n",x_range.max);
	ch_range.max=(*profile).be_status.data_ch;
	break;
      case UNDER_FLOW:
	//fprintf(stderr,"UnderFlow\n");
	ch_range.max=1;
	break;
      case OVER_FLOW:
	//fprintf(stderr,"OverFlow\n");
	ch_range.max=(*profile).be_status.data_ch;
	break;
      }
      //      fprintf(stderr," at RF=%f\n",x_range.max);
    }
  } else if(x_unit==X_BE_FREQ) {
    ch_range.min=(long)StarFreqOnBE2Ch(profile,&error_code,x_range.min);
    if(error_code!=NORMAL_END) {
      //      fprintf(stderr,"profile=%s:",(*profile).file_name);
      switch(error_code) {
      case UN_CONVERGE:
	fprintf(stderr,"profile=%s:",(*profile).file_name);
	fprintf(stderr,"Unconverge");
	fprintf(stderr," at BackendFreq=%f\n",x_range.min);
	ch_range.min=1;
	break;
      case UNDER_FLOW:
	//	fprintf(stderr,"UnderFlow");
	ch_range.min=1;
	break;
      case OVER_FLOW:
	//	fprintf(stderr,"OverFlow");
	ch_range.min=(*profile).be_status.data_ch;
	break;
      }
      //      fprintf(stderr," at BackendFreq=%f\n",x_range.min);
    }
    ch_range.max=(long)StarFreqOnBE2Ch(profile,&error_code,ceil(x_range.max));
    if(error_code!=NORMAL_END) {
      //      fprintf(stderr,"profile=%s:",(*profile).file_name);
      switch(error_code) {
      case UN_CONVERGE:
	fprintf(stderr,"profile=%s:",(*profile).file_name);
	fprintf(stderr,"Unconverge");
	fprintf(stderr," at BackendFreq=%f\n",x_range.max);
	ch_range.max=(*profile).be_status.data_ch;
	break;
      case UNDER_FLOW:
	//	fprintf(stderr,"UnderFlow\n");
	ch_range.max=1;
	break;
      case OVER_FLOW:
	//	fprintf(stderr,"OverFlow\n");
	ch_range.max=(*profile).be_status.data_ch;
	break;
      }
      //      fprintf(stderr," at BackendFreq=%f\n",x_range.max);
    }
  } else {
    ch_range.min=(long)StarVelocity2Ch(profile,&error_code,x_range.min);
    if(error_code!=NORMAL_END) {
      //      fprintf(stderr,"profile=%s:",(*profile).file_name);
      switch(error_code) {
      case UN_CONVERGE:
	fprintf(stderr,"profile=%s:",(*profile).file_name);
	fprintf(stderr,"Unconverge");
	fprintf(stderr," at velocity=%f\n",x_range.min);
	ch_range.min=1;
	break;
      case UNDER_FLOW:
	//	fprintf(stderr,"UnderFlow");
	ch_range.min=1;
	break;
      case OVER_FLOW:
	//	fprintf(stderr,"OverFlow");
	ch_range.min=(*profile).be_status.data_ch;
	break;
      }
      //      fprintf(stderr," at velocity=%f\n",x_range.min);
    }
    ch_range.max=(long)StarVelocity2Ch(profile,&error_code,ceil(x_range.max));
    if(error_code!=NORMAL_END) {
      //      fprintf(stderr,"profile=%s:",(*profile).file_name);
      switch(error_code) {
      case UN_CONVERGE:
	fprintf(stderr,"profile=%s:",(*profile).file_name);
	fprintf(stderr,"Unconverge");
	fprintf(stderr," at velocity=%f\n",x_range.max);
	ch_range.max=(*profile).be_status.data_ch;
	break;
      case UNDER_FLOW:
	//	fprintf(stderr,"UnderFlow\n");
	ch_range.max=1;
	break;
      case OVER_FLOW:
	//	fprintf(stderr,"OverFlow\n");
	ch_range.max=(*profile).be_status.data_ch;
	break;
      }
      //      fprintf(stderr," at velocity=%f\n",x_range.max);
    }
  }

  StarSortRangeLong(&ch_range);

  (*act_ch_range).min=ch_range.min;
  (*act_ch_range).max=ch_range.max;

  return(error_code);
}
