/*
STAR Profile data processing functions

1992/12/23 T.Handa
1992/12/28 T.Handa
1997/12/22 T.Handa ver.3
1999/ 8/13 T.Handa (debug) for ver.3 profile
1999/ 8/24 T.Handa (debug) StarFreqOnBE2Ch() the convergence criterion
2000/ 2/29 T.Handa (add)
2001/ 1/24 T.Handa (debug) StarFrequency2Ch() RF for LSB/USB
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>

#include <StData.h>
#include <StStd.h>

/*-----profile add/sub calcuration----*/

int StarSubtractProfile(out_prf,in1_prf,in2_prf)
/*
subtract two profiles
header of out_prf is not set from neither in1_prf nor in2_prf
out_prf=in1_prf-in2_prf;

return: =0 normal end, =999 different channel numbers between two inputs
1997/12/17 T.Handa
*/
PROFILE	*out_prf,	/* <out> out_prf=in1_prf-in2_prf */
  *in1_prf,	/* <in> */
  *in2_prf;	/* <in> */
{
  int ch,ii;
  int max_ch;
  double in1,in2,out,factor,f1,f2;
  int err_code;

  err_code=0;

/* scale if scaling factors are different */
  f1=(double)(*in1_prf).rx_status.scale;
  f2=(double)(*in2_prf).rx_status.scale;
  if(f1==f2)
    factor=f1;
  else
    factor=1.0;
  if((*in1_prf).be_status.data_ch!=(*in1_prf).be_status.data_ch)
    err_code=999;

/* do subtraction */
  max_ch=(*in1_prf).be_status.data_ch;
  for(ch=1;ch<=max_ch;ch++) {
    ii=ch-1;
    in1=(*in1_prf).data[ii]*f1;
    in2=(*in2_prf).data[ii]*f2;
    out=in1-in2;
    (*out_prf).data[ii]=(float)(out/factor);
  }

/* store the actual scale factor */
  (*out_prf).rx_status.scale=(float)factor;
  return(err_code);
}


int	StarAddProfile(out_prf,in1_prf,in2_prf)
/*
add two profiles
header of the out_prf is blank

out_prf=in1_prf+in2_prf;

return) =0 normal, =999 Two profiles have differcent channel numbers.
1997/12/17 T.Handa
*/
PROFILE	*out_prf,	/* <out> resultant added profile, out_prf=in1_prf+in2_prf */
	*in1_prf,	/* <in> */
	*in2_prf;	/* <in> */
{
  int ch,ii;
  int max_ch;
  double in1,in2,out,factor,f1,f2;
  int err_code;

  err_code=0;

/* If scaling factors are different, calculate in real values. */
  f1=(double)(*in1_prf).rx_status.scale;
  f2=(double)(*in2_prf).rx_status.scale;
  if(f1==f2)
    factor=f1;
  else
    factor=1.0;
  if((*in1_prf).be_status.data_ch!=(*in1_prf).be_status.data_ch)
    err_code=999;

/* make addition of profile.data */
  max_ch=(*in1_prf).be_status.data_ch;
  for(ch=1;ch<=max_ch;ch++) {
    ii=ch-1;
    in1=(*in1_prf).data[ii]*f1;
    in2=(*in2_prf).data[ii]*f2;
    out=in1+in2;
    (*out_prf).data[ii]=(float)(out/factor);
  }

/* store the actual cale factor */
  (*out_prf).rx_status.scale=(float)factor;
  return(err_code);
}

void	StarMultiplyProfile(out_prf,ff,in_prf)
/*
scalor production of profile
header of the out_prf is kept
out_prf=ff*in_prf;

return) =0 normal, =999 Two profiles have differcent channel numbers.
1997/12/17 T.Handa
2006/1/20 T.Handa (revise)
*/
PROFILE	*out_prf;	/* <out> resultant profile */
double	ff;	/* <in> scaling factor */
PROFILE	*in_prf;	/* <in> */
{
  int ch,ii;
  int max_ch;
  double in,out;

/* do scaling of profile.data */
  max_ch=(*in_prf).be_status.data_ch;
  for(ch=1;ch<=max_ch;ch++) {
    ii=ch-1;
    in=(double)(*in_prf).data[ii];
    out=in*ff;
    (*out_prf).data[ii]=(float)(out);
  }

  return;
}

int StarLinearCombProfile(out_prf,ff1,in1_prf,ff2,in2_prf)
/*
make a linear combination of two profiles
header of the out_prf is blank
out_prf=ff1*in1_prf+ff2*in2_prf;

return) =0 normal, =999 Two profiles have differcent channel numbers.
1997/12/22 T.Handa
*/
PROFILE	*out_prf;	/* <out> resultant profile */
double	ff1;	/* <in> scaling factor for in1_prf */
PROFILE	*in1_prf;	/* <in> */
double	ff2;	/* <in> scaling factor for in2_prf */
PROFILE	*in2_prf;	/* <in> */
{
  int ch,ii;
  int max_ch;
  double in1,in2,out,factor,f1,f2;
  int err_code;

  err_code=0;

/* If scaling factors are different, calculate in real values. */
  f1=(double)(*in1_prf).rx_status.scale;
  f2=(double)(*in2_prf).rx_status.scale;
  if(f1==f2)
    factor=f1;
  else
    factor=1.0;
  if((*in1_prf).be_status.data_ch!=(*in1_prf).be_status.data_ch)
    err_code=999;

/* make a linear combination of profile.data */
  max_ch=(*in1_prf).be_status.data_ch;
  for(ch=1;ch<=max_ch;ch++) {
    ii=ch-1;
    in1=(*in1_prf).data[ii]*f1;
    in2=(*in2_prf).data[ii]*f2;
    out=ff1*in1+ff2*in2;
    (*out_prf).data[ii]=(float)(out/factor);
  }

/* store the actual cale factor */
  (*out_prf).rx_status.scale=(float)factor;
  return(err_code);
}


/*-----Frequency and Velocity calcuration----*/

/* definitions for iteration in StarFreqOnBECh(profile,freq) */ 
#define	POINTS	5	/* interation start points, minimum 2 */
#define	LOOP_LIMIT	100	/* interation loop limit */


double StarBindedCh(profile,ch)
/*
estimate the channel position on the original backend from that on data array 

The actual data is stored between 1 to profile.be_status.data_ch in data array,
and between 1 to profile.be_status.be_ch in the original backend.
The channel frequency is at the center of the channel.

1992/12/27 T.Handa
1999/ 8/13 T.Handa (debug) ineffective for ver.3 profile
*/
PROFILE	*profile; /* <in> */
double ch; /* <in> */
{
  double ch_data;

  if((*profile).sys_head.version>2) {
    return ch;
  }

/* assume the current data are sampled at smoothing boxcar width. */
  ch_data=(ch-(double)(*profile).be_status.ch_orig+0.5)/(double)(*profile).be_status.ch_width+0.5;

  return ch_data;
}

double StarBEDeltaFreqOnBE(profile,ch)
/*
estimate the derivative coefficient of the backend dispersion parameters

The actual data is stored between 1 to profile.be_status.be_ch 
in the original backend.

1992/12/27 T.Handa
1993/6/16 Naomi Hirano (revise the spec & debug)
1996/1/11 T.Handa (revise) return the spec to the older one
1996/1/11 T.Handa (revise) available also for decreasing ch-freq relation
1999/ 8/13 T.Handa (debug) ineffective for ver.3 profile
*/
PROFILE *profile; /* <in> */
double ch; /* <in> */
{
  double delta_freq,	/* freq on the Backend(MHz) */
    ch_pos;

  if((*profile).sys_head.version>2) {
    ch_pos=ch-(*profile).be_status.center_ch;
  } else {
    ch_pos=StarOriginalCh(profile,ch)-(*profile).be_status.basis_ch;
  }

  delta_freq=(((*profile).be_status.disp4*ch_pos*4.0+(*profile).be_status.disp3)*ch_pos*3.0+(*profile).be_status.disp2)*ch_pos*2.0+(*profile).be_status.disp1;

  return(delta_freq);
}

double StarFreqOnBE2Ch(profile, error_code, freq)
/*
estimate the channel position on the backend from frequency on the backend.
(ver.3 or later, channel position on the data array)

The actual data is stored between 1 to profile.be_status.data_ch in data array,
and between 1 to profile.be_status.be_ch in the original backend.
The channel frequency is at the center of the channel.

1992/12/28 T.Handa
1993/6/16 Naomi Hirano (debug)
1996/1/11 T.Handa (revise) available also for decreasing ch-freq relation
1996/1/24 T.Handa (revise) add the parameter "error_code"
*/
PROFILE *profile; /* <in> */
int *error_code; /* <out> */
double freq; /* <in> freq on the backend (MHz) */
{
  double freq_limit, freq_resid, freq_iter[2], ch_iter[2];
  double ch_points[POINTS],freq_points[POINTS];
  double neighbor_ch[2],neighbor_freq[2];
  double freq_dev;

  int ii;

  long total_ch;

  if((*profile).sys_head.version<3)
    total_ch=(long)(*profile).be_status.be_ch;
  else
    total_ch=(long)(*profile).be_status.data_ch;

/* iteration limit = 1/20 of the resolution */
  freq_limit=(*profile).be_status.freq_res*0.05;

/* calculate standard points */ 
  for(ii=0;ii<POINTS;ii++) {
    ch_points[ii]=(double)(total_ch/(POINTS-1)*ii+1);
  }
  ch_points[0]=(-1.0); ch_points[POINTS-1]=ch_points[POINTS-1]+1.0;
  if(ch_points[POINTS-1]<(double)(total_ch+1))
    ch_points[POINTS-1]=(double)(total_ch+1);

  for(ii=0;ii<POINTS;ii++) {
    freq_points[ii]=StarCh2FreqOnBE(profile,ch_points[ii]);
  }

/* out of the BE coverage */
/* increasing freq on ch */
  if(freq_points[0]<freq_points[1]){
    if(freq>freq_points[POINTS-1]) {
      *error_code=OVER_FLOW;
      return(freq_points[POINTS-1]);
    }
    else if(freq<freq_points[0]) {
      *error_code=UNDER_FLOW;
      return(freq_points[0]);
    }
  }
/* decreasing freq on ch */
  else{
    if(freq<freq_points[POINTS-1]) {
      *error_code=UNDER_FLOW;
      return(freq_points[POINTS-1]);
    }
    else if(freq>freq_points[0]) {
      *error_code=OVER_FLOW;
      return(freq_points[0]);
    }
  }

/* find the nearest two standard points */
/* increasing freq on ch */
  if(freq_points[0]<freq_points[1]){
    for(ii=0;ii<POINTS;ii++) {
      if(freq_points[ii]<=freq) {
	neighbor_ch[0]=ch_points[ii];
	neighbor_freq[0]=freq_points[ii];
      }
    }
    for(ii=POINTS-1;ii>=0;ii--) {
      if(freq_points[ii]>=freq) {
	neighbor_ch[1]=ch_points[ii];
	neighbor_freq[1]=freq_points[ii];
      }
    }
  }
/* decreasing freq on ch */
  else{
    for(ii=0;ii<POINTS;ii++) {
      if(freq_points[ii]>=freq) {
	neighbor_ch[0]=ch_points[ii];
	neighbor_freq[0]=freq_points[ii];
      }
    }
    for(ii=POINTS-1;ii>=0;ii--) {
      if(freq_points[ii]<=freq) {
	neighbor_ch[1]=ch_points[ii];
	neighbor_freq[1]=freq_points[ii];
      }
    }
  }
/* hit the standard point */
  if(neighbor_freq[0]==neighbor_freq[1]) {
    ch_iter[0]=StarBindedCh(profile,ch_iter[0]);
    *error_code=NORMAL_END;
    return(ch_iter[0]);
  }

/* linear evaluation */
  ch_iter[0]=(freq-neighbor_freq[0])/(neighbor_freq[1]-neighbor_freq[0])
    *(neighbor_ch[1]-neighbor_ch[0])+neighbor_ch[0];
  freq_iter[0]=StarCh2FreqOnBE(profile,ch_iter[0]);

/* iteration by Newton-Lapson method */
  for(ii=0;ii<LOOP_LIMIT;ii++) {
    freq_dev=StarBEDeltaFreqOnBE(profile,ch_iter[0]);
    ch_iter[1]=ch_iter[0]-(freq_iter[0]-freq)/freq_dev;
    freq_iter[1]=StarCh2FreqOnBE(profile,ch_iter[1]);

    freq_resid=freq_iter[1]-freq;
    if(fabs(freq_resid)<=freq_limit) break;

    ch_iter[0]=ch_iter[1];
    freq_iter[0]=freq_iter[1];
  }

/* check conversion */
  if(fabs(freq_resid)>freq_limit){
    *error_code=UN_CONVERGE;
    return(ch_iter[0]);
  }

/* change dispersion parameter definition after ver.3 */
  if((*profile).sys_head.version<3){
    ch_iter[1]=StarBindedCh(profile,ch_iter[1]);
  }
  *error_code=NORMAL_END;
  return(ch_iter[1]);	
}

double StarFrequency2Ch(profile,error_code,freq)
/*
estimate the channel position on the data array from frequency.

The actual data is stored between 1 to profile.be_status.data_ch in data array.
The channel frequency is at the center of the channel.

1992/12/27 T.Handa
1996/ 1/24 T.Handa (revise) add "error_code", change "rest_freq" to "obs_freq"
2001/ 1/24 T.Handa (debug) RF for LSB/USB
2006/ 2/ 7 T.Handa (debug) LSB/USB ver.5
*/
PROFILE *profile; /* <in> */
int	*error_code; /* <in> */
double freq;	/* RF frequency (GHz) */
{
  double be_freq, /* freq on the backend (MHz) */
    center_freq, /* nominal center_freq on the backend (MHz) */
    offset_freq; /* offset from cetner_freq (MHz) */
  double data_ch;

  center_freq=(*profile).be_status.center_freq;
/*  offset_freq=(freq-(*profile).rx_status.rest_freq)*1.0e3;*/
  offset_freq=(freq-(*profile).rx_status.obs_freq)*1.0e3;

  if((*profile).be_status.sign<0)
    offset_freq= -offset_freq;

  if((*profile).sys_head.version<=4) { // older version
    if((*profile).rx_status.sb_mode==LSB_SSB || (*profile).rx_status.sb_mode==LSB_DSB) 
      offset_freq= -offset_freq;
  }

/* adjust for VST2 */
  if(strncmp((*profile).ant_status.tele,"VST2 ",5)==0)
    offset_freq= -offset_freq;

  be_freq=center_freq+offset_freq;

  data_ch=StarFreqOnBE2Ch(profile,error_code,be_freq);

  return(data_ch);
}


double StarVelocity2Ch(profile,error_code,v_lsr)
/*
estimate the channel position on the data array from velocity.

The actual data is stored between 1 to profile.be_status.data_ch in data array.
The channel velocity is at the center of the channel.

1992/12/28 T.Handa
1996/ 1/24 T.Handa (revise) add "error_code", change "rest_freq" to "obs_freq"
*/
PROFILE *profile; /* <in> */
int	*error_code; /* <out> */
double v_lsr;	/* <in> (km/s) */
{
  double offset_vel, /* offset from tracking velocity (km/s) */
    offset_freq, /* offset from tracking freq (GHz) */
    rf_freq, /* RF freq (GHz) */
    data_ch;

  offset_vel=v_lsr-(*profile).rx_status.v_lsr;
/* frequency for redshift calculation (old) */
/*  offset_freq=(-offset_vel)*((*profile).rx_status.rest_freq/LIGHT_SPEED);

  rf_freq=offset_freq+(*profile).rx_status.rest_freq;*/
/* frequency for redshift calculation (new) */
  offset_freq=(-offset_vel)*((*profile).rx_status.obs_freq/LIGHT_SPEED);

  rf_freq=offset_freq+(*profile).rx_status.obs_freq;

  data_ch=StarFrequency2Ch(profile,error_code,rf_freq);

  return(data_ch);
}


double StarOriginalCh(profile,ch)
/*
estimate the channel position on data array from that on the original backend 

The actual data is stored between 1 to profile.be_status.data_ch in data array,
and between 1 to profile.be_status.be_ch in the original backend.
The channel frequency is at the center of the channel.

1992/12/25 T.Handa
1999/ 8/13 T.Handa (debug) ineffective for ver.3 profile
*/
PROFILE	*profile;
double ch;
{
  double	ch_be;

  if((*profile).sys_head.version>2) {
    return ch;
  }
/* assume the current data are sampled at smoothing boxcar width. */
  ch_be	=(ch-0.5)*(double)(*profile).be_status.ch_width+(double)(*profile).be_status.ch_orig-0.5;

  return ch_be;
}

double StarCh2FreqOnBE(profile,ch)
/*
estimate the channel position on the original backend from frequency on the backend.

The actual data is stored between 1 to profile.be_status.data_ch in data array,
and between 1 to profile.be_status.be_ch in the original backend.
The channel frequency is at the center of the channel.

1992/12/25 T.Handa
1999/ 8/13 T.Handa (debug) ineffective for ver.3 profile
*/
PROFILE *profile; /* <in> */
double ch; /* <in> */
{
  double be_freq, /* freq on the backend (MHz) */
    ch_pos;

  if((*profile).sys_head.version>2) {
    ch_pos=ch-(*profile).be_status.center_ch;
  } else {
    ch_pos=StarOriginalCh(profile,ch)-(*profile).be_status.basis_ch;
  }

  be_freq=((((*profile).be_status.disp4*ch_pos+(*profile).be_status.disp3)*ch_pos+(*profile).be_status.disp2)*ch_pos+(*profile).be_status.disp1)*ch_pos+(*profile).be_status.disp0;

  return(be_freq);
}

double StarCh2Frequency(profile,ch)
/*
estimate obs freq at ch for data
First data is profile.data[0] at ch=1
Last data is profile.data[ch-1] at ch=profile.be_status.data_ch

1992/12/27 T.Handa
1996/1/24 T.Handa (debug) center freq corresponds not to rest_freq but obs_freq
2001/ 1/24 T.Handa (debug) RF for LSB/USB
2006/ 2/ 7 T.Handa (debug) LSB/USB ver.5
*/
PROFILE *profile;
double ch;
{
  double be_freq,	/* frequency on the backend(MHz) */
    center_freq,	/* frequency of the "center" of the backend(MHz) */
    offset_freq,	/* frequency offset from the "center"(MHz) */
    rf_freq;	/* receiving frequency(GHz) */

  be_freq=StarCh2FreqOnBE(profile,ch);
  center_freq=(*profile).be_status.center_freq;

  offset_freq=be_freq-center_freq;

  if((*profile).be_status.sign<0)
    offset_freq= -offset_freq;

  if((*profile).sys_head.version<=4) { // older version
    if((*profile).rx_status.sb_mode==LSB_SSB || (*profile).rx_status.sb_mode==LSB_DSB) 
      offset_freq= -offset_freq;
  }

/* adjust for VST2 */
  if(strncmp((*profile).ant_status.tele,"VST2 ",5)==0)
    offset_freq= -offset_freq;

/*  rf_freq=offset_freq/1.0e3+(*profile).rx_status.rest_freq;*/
  rf_freq=offset_freq/1.0e3+(*profile).rx_status.obs_freq;

  return(rf_freq);
}
	
double StarCh2Velocity(profile,ch)
/*
estimate LSR velocity at ch for data
First data is profile.data[0] at ch=1
Last data is profile.data[ch-1] at ch=profile.be_status.data_ch

1992/12/28 T.Handa
1996/1/24 T.Handa (debug) center freq corresponds not to rest_freq but obs_freq
*/
PROFILE *profile;
double ch;
{
  double rf_freq,	/* receiving frequency(GHz) */
    offset_freq,	/* offset freq from the tracking freq(GHz) */
    offset_vel,	/* offset velocity from the tracking velocity(km/s) */
    velocity;	/* LSR velocity(km/s) */

  rf_freq=StarCh2Frequency(profile,ch);

/* calculate the redshift */
/*  offset_freq=rf_freq-(*profile).rx_status.rest_freq;
    offset_vel=(-offset_freq)/(*profile).rx_status.rest_freq*LIGHT_SPEED;*/

/* calculate the redshift */
  offset_freq=rf_freq-(*profile).rx_status.obs_freq;
  offset_vel=(-offset_freq)/(*profile).rx_status.obs_freq*LIGHT_SPEED;

  velocity=offset_vel+(*profile).rx_status.v_lsr;

  return(velocity);
}

/*-----BitMap processing----*/

void StarRReadBaselineFlags(profile,n_range,base_range)
/*
get ranges where baseline flag is on
1997/12/25 T.Handa
*/
PROFILE *profile;	/* <in> profile */
int *n_range;	/* <out> the number of ranges */
RANGE_long *base_range;	/* <out> ranges where baseline flag is on */
{
  int ch,max_ch;
  int ii,baseline_flag;
  int inout;

  max_ch=(*profile).be_status.data_ch;
  ii=0;
  inout= -1;

  for(ch=1;ch<=max_ch;ch++) {
    baseline_flag=StarReadBaselineFlag(profile,ch);
    if(baseline_flag==FLAGGED && inout<0) {
      base_range[ii].min=ch;
      inout= -inout;
    } else if(baseline_flag==UNFLAGGED && inout>0) {
      base_range[ii].max=ch-1;
      inout= -inout;
      ii++;
    }
  }
  *n_range=ii;

  return;
}

void StarRReadSpuriousFlags(profile,n_range,spur_range)
/*
get ranges where spurious flag is on
1997/12/25 T.Handa
*/
PROFILE *profile;	/* <in> profile */
int *n_range;	/* <out> the number of ranges */
RANGE_long *spur_range;	/* <out> ranges where spurious flag is on */
{
  int ch,max_ch;
  int ii,spurious_flag;
  int inout;

  max_ch=(*profile).be_status.data_ch;
  ii=0;
  inout= -1;

  for(ch=1;ch<=max_ch;ch++) {
    spurious_flag=StarReadSpuriousFlag(profile,ch);
    if(spurious_flag==FLAGGED && inout<0) {
      spur_range[ii].min=ch;
      inout= -inout;
    } else if(spurious_flag==UNFLAGGED && inout>0) {
      spur_range[ii].max=ch-1;
      inout= -inout;
      ii++;
    }
  }
  *n_range=ii;

  return;
}

void StarProfileClearBaselineFlag(profile,flag)
PROFILE *profile; /* <i/o> */
int flag; /* <in> */
/*
clear baseline flags of the baseline bitmap over all channels
1997/12/22 Osamu Kimura
*/
{
  RANGE_long ch_range;	/* range */ 

  ch_range.min=1;
  ch_range.max=(*profile).be_status.data_ch;

  StarRWriteBaselineFlags(profile,ch_range,flag);

  return;
}

void StarRWriteBaselineFlags(profile,ch_range,flag)
/*
write baseline flags of the baseline bitmap over an asigned channel range
1993/1/9 Osamu Kimura
*/
PROFILE		*profile; /* <i/o> */
RANGE_long	ch_range; /* <in> */
int		flag; /* <in> */
{
  int data_ch;

  data_ch = (*profile).be_status.data_ch;
  StarRWriteBitMapFlags(&((*profile).baseline[0]),ch_range,data_ch,flag);

  return;
}

void StarWriteBaselineFlag(profile,ch,flag)
/*
write baseline flags of the baseline bitmap at an asigned channel
1993/1/9 Osamu Kimura
*/
PROFILE		*profile; /* <i/o> */
int		ch;	/* <in> */ 
int		flag; /* <in> */
{
  int data_ch;

  data_ch = (*profile).be_status.data_ch;

  if(ch<1)
    ch = 1;
  else if(ch>data_ch)
    ch = data_ch;

  StarWriteBitMapFlag(&((*profile).baseline[0]),ch,data_ch,flag);

  return;
}	

int StarReadBaselineFlag(profile,ch)
/*
read baseline flag at an assigned channel
1993/1/9 Osamu Kimura
*/
PROFILE	*profile; /* <in> */
int ch; /* <in> */
{
  int flagbit;

  flagbit = StarReadBitMapFlag(ch,&((*profile).baseline[0]));
  return(flagbit);
}	


void StarProfileClearSpuriousFlag(profile,flag)
PROFILE *profile; /* <i/o> */
int flag; /* <in> */
/*
clear spurious flags of the baseline bitmap over all channels
1997/12/22 Osamu Kimura
*/
{
  RANGE_long ch_range;

  ch_range.min=1;
  ch_range.max=(*profile).be_status.data_ch;

  StarRWriteSpuriousFlags(profile,ch_range,flag);

  return;
}

void StarRWriteSpuriousFlags(profile,ch_range,flag)
/*
write spurious flags of the spurious bitmap over an asigned channel range
1993/1/9 Osamu Kimura
*/
PROFILE		*profile; /* <i/o> */
RANGE_long	ch_range; /* <in> */ 
int		flag; /* <in> */
{
  int data_ch;

  data_ch = (*profile).be_status.data_ch;
  StarRWriteBitMapFlags(&((*profile).spurious[0]),ch_range,data_ch,flag);

  return;
}

void StarWriteSpuriousFlag(profile,ch,flag)
/*
write spurious flags of the spurious bitmap at an asigned channel
1993/1/9 Osamu Kimura
*/
PROFILE *profile; /* <i/o> */
int ch;	/* <in> */ 
int flag; /* <in> */
{
  int data_ch;

  data_ch = (*profile).be_status.data_ch;

  if(ch<1)
    ch = 1;
  else if(ch>data_ch)
    ch = data_ch;

  StarWriteBitMapFlag(&((*profile).spurious[0]),ch,data_ch,flag);

  return;
}

int StarReadSpuriousFlag(profile,ch)
/*
read spurious flag at an assigned channel
1993/1/9 Osamu Kimura
*/
PROFILE	*profile;
int ch;
{
  int flagbit;

  flagbit = StarReadBitMapFlag(ch,&((*profile).spurious[0]));
  return(flagbit);
}

int StarReadBitMapFlag (ch, bitmap_flag)
/*
read BitMap Flag (for baseline and/or spurious)
return) FLAGGED, UNFLAGGED
1992 Dec. 18 Written by Neko
*/
int ch;	/* <in> */
unsigned char bitmap_flag[];	/* <in> */

{
  int byte_number, bit_number;	/* location on bitmap array */
  int flag;
  static unsigned char mask = 0x01;

/* estimate address in the bitmap array */

  byte_number = (ch - 1) / (sizeof(unsigned char)* 8) ;
  bit_number = ch - byte_number*sizeof(unsigned char)* 8 - 1;

/* check ON/OFF of flag */
	
  if ((bitmap_flag[byte_number] & mask << bit_number) != 0) {
    flag = FLAGGED;
  } else {
    flag = UNFLAGGED;
  }

  return (flag);
}

void StarWriteBitMapFlag(bitmap_flag, ch, data_ch, flag)
/*
write flags on the bitmap at an assigned channel
1992 Dec. 18 Written by Neko
1993/1/8 T.Handa
*/
unsigned char bitmap_flag[];	/* <in> bitmap for flag */
int ch;		/* <in> */
int data_ch;	/* <in> last channel of bitmap */
int flag;	/* <in> FLAGGED,UNFLAGGED */
{
  int byte_number, bit_number;
  unsigned char	flag_byte;
  static unsigned char	mask=0x01;

/* ch adjust */
  if(ch<1)
    ch=1; 
  if(ch>data_ch) 
    ch=data_ch;

/* calculate location */
  byte_number = (ch - 1) / (sizeof(unsigned char)* 8) ;
  bit_number = ch - byte_number*sizeof(unsigned char)* 8 - 1;

/* set flag */
  if(flag==0){
    flag_byte=~bitmap_flag[byte_number];
    flag_byte=flag_byte | mask << bit_number;
    bitmap_flag[byte_number]=~flag_byte;
  }
  else{
    flag_byte=bitmap_flag[byte_number];
    flag_byte=flag_byte | mask << bit_number;
    bitmap_flag[byte_number]=flag_byte;
  }

  return;
}

void StarRWriteBitMapFlags(bitmap_flag, ch_range, data_ch, flag)
/*
write flags on the bitmap over an assigned channel range
1992 Dec. 18 Written by Neko
1993/1/8 T.Handa
*/
unsigned char bitmap_flag[];	/* <out> bitmap */
RANGE_long ch_range;		/* <in> (sort internally) */
int data_ch;			/* <in> last ch on bitmap */
int flag;			/* <in> FLAGGED,UNFLAGGED */
{
  int flag_ch;
  int byte_number;

  unsigned char	flag_byte;

  flag_ch = (data_ch - 1)/(sizeof(unsigned char)*8) +1;
	
/* sort ch_range */
  StarSortRangeLong(&ch_range);

/* set flags in ch_range */

  {
    int ch = 1;
    int fswitch;
    /* fswitch = 0 : before the range */
    /* fswitch = 1 : in the range */
    /* fswitch = 2 : after the range */

    fswitch = 0;

    for (byte_number = 0; byte_number < flag_ch; ++byte_number) {
      int bit_number;
      unsigned char mask = 1;

      for (bit_number = 0; bit_number < 8; ++bit_number) {
	if (fswitch == 0) {
	  if (ch >= ch_range.min) {
	    fswitch = 1;
	    if(flag==0){
	      flag_byte=~bitmap_flag[byte_number];
	      flag_byte=flag_byte | mask << bit_number;
	      bitmap_flag[byte_number]=~flag_byte;
	    } else {
	      bitmap_flag[byte_number] = bitmap_flag[byte_number] | mask << bit_number;
	    }
	  }
	} else {
	  if (ch <= ch_range.max) {
	    if(flag==0){
	      flag_byte=~bitmap_flag[byte_number];
	      flag_byte=flag_byte | mask << bit_number;
	      bitmap_flag[byte_number]=~flag_byte;
	    } else {
	      bitmap_flag[byte_number] = bitmap_flag[byte_number] | mask << bit_number;
	    }
	  } else {
	    fswitch = 2;
	  }
	}
/*        printf ("ch = %d fswitch = %d\n", ch, fswitch); */
	++ch;
      }
    }
  }

  return;
}

int	StarAddHistory(profile,history)
/*
add HISTORY to profile.
END is automatically attached in the last keyword
END itself can be attached by this function.

return)
0   normal end
100 history step buffer overflow

1992/12/12 T.Handa
*/
PROFILE	*profile;	/* <i/o> */
HISTORY	*history;	/* <in> to be added */
{
  int	lines;
  long	pres_time;
  int	current_step;
  int	add_size,add_byte;

  current_step=(*profile).sys_head.hist_step;

/* ----correct increasing the sizes---- */
/* previous last step is "END" --> adjust sizes */
  if(strcmp((*profile).history[current_step].step,"END")==0 ||
     strcmp((*profile).history[current_step].step,"end")==0) {
    (*profile).sys_head.hist_line--;
    (*profile).sys_head.l_file-=16;
    (*profile).sys_head.l_head-=16;
  }

  add_size=(*history).size+1;
  add_byte=add_size*16;

  (*profile).sys_head.hist_line+=add_size;
  (*profile).sys_head.l_file+=add_byte;
  (*profile).sys_head.l_head+=add_byte;

/* increase history step in normal history addition */
  if(strcmp((*history).step,"END")!=0
     && strcmp((*history).step,"end")!=0) {
    (*profile).sys_head.hist_step++;
  } else {
    (*history).size=0;
  }

/* too many history steps */
  if((*profile).sys_head.hist_step>=MAX_HISTORY_STEP)
    return(100);

/* copy history to profile.history */
  (*profile).history[current_step].size=(*history).size;
  (*history).step[7]=(char)NULL;
  strcpy((*profile).history[current_step].step,(*history).step);
  time(&pres_time);
  (*profile).history[current_step].time=pres_time;
  for(lines=0;lines<(*history).size;lines++) {
    (*history).prm[lines][16]=(char)NULL;
    strcpy((*profile).history[current_step].prm[lines],(*history).prm[lines]);
  }

/* add history is alread "END" */
  if(strcmp((*history).step,"END")==0
     || strcmp((*history).step,"end")==0)
    return(0);

/* ---add "END" step--- */

  (*profile).sys_head.hist_line++;
  (*profile).sys_head.l_file+=16;
  (*profile).sys_head.l_head+=16;
  current_step=(*profile).sys_head.hist_step;
/* too many history steps */
  if((*profile).sys_head.hist_step>=MAX_HISTORY_STEP)
    return(100);

  (*profile).history[current_step].size=0;
  strcpy((*profile).history[current_step].step,"END");	
  (*profile).history[current_step].time=pres_time;

  return(0);
}

/* -------PROFILE data processing -------*/

void	StarCopyProfile(out_prf,in_prf)
/*
copy profile
1992/12/11 T.Handa
*/
PROFILE	*out_prf,	/* <out> */
	*in_prf;	/* <in> */
{
  int ii;
  unsigned short history_steps;

/* copy STAR pipeflow header */
  strcpy((*out_prf).file_id,(*in_prf).file_id);
  strcpy((*out_prf).file_name,(*in_prf).file_name);

/* copy STAR file system header */
  StarCopySyshead(&((*out_prf).sys_head),&((*in_prf).sys_head));
  history_steps=(unsigned short)(*out_prf).sys_head.hist_step;

/* copy scan file header */
  StarCopyObsprm(&((*out_prf).obs_prm),&((*in_prf).obs_prm));
  StarCopyObject(&((*out_prf).object),&((*in_prf).object));
  StarCopyAntstatus(&((*out_prf).ant_status),&((*in_prf).ant_status));
  StarCopyRxstatus(&((*out_prf).rx_status),&((*in_prf).rx_status));
  StarCopyBestatus(&((*out_prf).be_status),&((*in_prf).be_status));
  StarCopyObslog(&((*out_prf).obs_log),&((*in_prf).obs_log));
/* copy history */
  for(ii=0;ii<=history_steps;ii++) {
    StarCopyHistory(&((*out_prf).history[ii]),&((*in_prf).history[ii]));
  }
/* copy spurious bitmap */
  StarCopySpurious((*out_prf).be_status.data_ch,(*out_prf).spurious,(*in_prf).spurious);
/* copy baseline bitmap */
  StarCopyBaseline((*out_prf).be_status.data_ch,(*out_prf).baseline,(*in_prf).baseline);
/* copy profile data */
  StarCopyProfdata((*out_prf).be_status.data_ch,(*out_prf).data,(*in_prf).data);

  return;
}


/* --- Part for StarCopyProfile --- */
void	StarCopySyshead(out_sys_head,in_sys_head)
SYS_HEAD	*out_sys_head,*in_sys_head;
{
  (*out_sys_head).l_file=(*in_sys_head).l_file;
  (*out_sys_head).l_head=(*in_sys_head).l_head;
  (*out_sys_head).format_id=(*in_sys_head).format_id;
  (*out_sys_head).version=(*in_sys_head).version;
  (*out_sys_head).data_type=(*in_sys_head).data_type;
  (*out_sys_head).hist_step=(*in_sys_head).hist_step;
  (*out_sys_head).hist_line=(*in_sys_head).hist_line;
  (*out_sys_head).scan=(*in_sys_head).scan;
  (*out_sys_head).backup=(*in_sys_head).backup;
  (*out_sys_head).num_scan=(*in_sys_head).num_scan;

  return;
}
void	StarCopyObject(out_object,in_object)
OBJECT	*out_object,*in_object;
{
  strcpy((*out_object).name,(*in_object).name);
  (*out_object).cood_sys=(*in_object).cood_sys;
  (*out_object).on_lb.x=(*in_object).on_lb.x;
  (*out_object).on_lb.y=(*in_object).on_lb.y;
  (*out_object).on_radec.x=(*in_object).on_radec.x;
  (*out_object).on_radec.y=(*in_object).on_radec.y;
  strcpy((*out_object).equinox,(*in_object).equinox);
  (*out_object).on_azel.x=(*in_object).on_azel.x;
  (*out_object).on_azel.y=(*in_object).on_azel.y;
  (*out_object).off_lb.x=(*in_object).off_lb.x;
  (*out_object).off_lb.y=(*in_object).off_lb.y;
  (*out_object).off_radec.x=(*in_object).off_radec.x;
  (*out_object).off_radec.y=(*in_object).off_radec.y;
  (*out_object).orig_xy.x=(*in_object).orig_xy.x;
  (*out_object).orig_xy.y=(*in_object).orig_xy.y;
  (*out_object).pa_xy=(*in_object).pa_xy;
  (*out_object).on_xy.x=(*in_object).on_xy.x;
  (*out_object).on_xy.y=(*in_object).on_xy.y;
  (*out_object).azel_offset.x=(*in_object).azel_offset.x;
  (*out_object).azel_offset.y=(*in_object).azel_offset.y;

  return;
}
void	StarCopyObsprm(out_obs_prm,in_obs_prm)
OBS_PRM	*out_obs_prm,*in_obs_prm;
{
  strcpy((*out_obs_prm).project,(*in_obs_prm).project);
  strcpy((*out_obs_prm).scd_file,(*in_obs_prm).scd_file);
  (*out_obs_prm).line_numb=(*in_obs_prm).line_numb;
  (*out_obs_prm).sw_mode=(*in_obs_prm).sw_mode;
  (*out_obs_prm).mjd=(*in_obs_prm).mjd;
  (*out_obs_prm).start=(*in_obs_prm).start;
  (*out_obs_prm).stop=(*in_obs_prm).stop;
  (*out_obs_prm).on_time=(*in_obs_prm).on_time;
  (*out_obs_prm).off_time=(*in_obs_prm).off_time;
  strcpy((*out_obs_prm).observer1,(*in_obs_prm).observer1);
  strcpy((*out_obs_prm).observer2,(*in_obs_prm).observer2);
  strcpy((*out_obs_prm).observer3,(*in_obs_prm).observer3);
  strcpy((*out_obs_prm).observer4,(*in_obs_prm).observer4);

  return;
}
void	StarCopyAntstatus(out_ant_status,in_ant_status)
ANT_STATUS	*out_ant_status,*in_ant_status;
{
  strcpy((*out_ant_status).tele,(*in_ant_status).tele);
  (*out_ant_status).ap_eff=(*in_ant_status).ap_eff;
  (*out_ant_status).mb_eff=(*in_ant_status).mb_eff;
  (*out_ant_status).fss_eff=(*in_ant_status).fss_eff;
  (*out_ant_status).hpbw=(*in_ant_status).hpbw;

  return;
}
void	StarCopyRxstatus(out_rx_status,in_rx_status)
/*
1999/ 8/13 T.Handa (debug) for ver.3 profile
*/
RX_STATUS	*out_rx_status,*in_rx_status;
{
  strcpy((*out_rx_status).name,(*in_rx_status).name);
  (*out_rx_status).sb_mode=(*in_rx_status).sb_mode;
  (*out_rx_status).v_lsr=(*in_rx_status).v_lsr;
  (*out_rx_status).v_off=(*in_rx_status).v_off;
  (*out_rx_status).rest_freq=(*in_rx_status).rest_freq;
  (*out_rx_status).obs_freq=(*in_rx_status).obs_freq;
  (*out_rx_status).lo_freq1=(*in_rx_status).lo_freq1;
  (*out_rx_status).if_freq1=(*in_rx_status).if_freq1;
  (*out_rx_status).lo_freq2=(*in_rx_status).lo_freq2;
  (*out_rx_status).if_freq2=(*in_rx_status).if_freq2;
  (*out_rx_status).lo_freq3=(*in_rx_status).lo_freq3;
  (*out_rx_status).if_freq3=(*in_rx_status).if_freq3;
  (*out_rx_status).mltplx1=(*in_rx_status).mltplx1;
  (*out_rx_status).mltplx2=(*in_rx_status).mltplx2;
  (*out_rx_status).mltplx3=(*in_rx_status).mltplx3;
  (*out_rx_status).tsys=(*in_rx_status).tsys;
  (*out_rx_status).trx=(*in_rx_status).trx;
  (*out_rx_status).rms=(*in_rx_status).rms;
  (*out_rx_status).calib=(*in_rx_status).calib;
  (*out_rx_status).weight=(*in_rx_status).weight;
  (*out_rx_status).scale=(*in_rx_status).scale;

  return;
}
void	StarCopyBestatus(out_be_status,in_be_status)
/*
1999/ 8/13 T.Handa (debug) for ver.3 profile
*/
BE_STATUS	*out_be_status,*in_be_status;
{
  strcpy((*out_be_status).name,(*in_be_status).name);
  (*out_be_status).be_ch=(*in_be_status).be_ch;
  (*out_be_status).basis_ch=(*in_be_status).basis_ch;
  (*out_be_status).sign=(*in_be_status).sign;
  (*out_be_status).data_ch=(*in_be_status).data_ch;
  (*out_be_status).center_ch=(*in_be_status).center_ch;
  (*out_be_status).center_freq=(*in_be_status).center_freq;
  (*out_be_status).disp0=(*in_be_status).disp0;
  (*out_be_status).disp1=(*in_be_status).disp1;
  (*out_be_status).disp2=(*in_be_status).disp2;
  (*out_be_status).disp3=(*in_be_status).disp3;
  (*out_be_status).disp4=(*in_be_status).disp4;
  (*out_be_status).freq_res=(*in_be_status).freq_res;
  (*out_be_status).ch_width=(*in_be_status).ch_width;
  (*out_be_status).ch_orig=(*in_be_status).ch_orig;

  return;
}
void	StarCopyObslog(out_obs_log,in_obs_log)
OBS_LOG	*out_obs_log,*in_obs_log;
{
  strcpy((*out_obs_log).site_name,(*in_obs_log).site_name);
  (*out_obs_log).meteo.tmp=(*in_obs_log).meteo.tmp;
  (*out_obs_log).meteo.press=(*in_obs_log).meteo.press;
  (*out_obs_log).meteo.water=(*in_obs_log).meteo.water;
  (*out_obs_log).tau0=(*in_obs_log).tau0;
  strcpy((*out_obs_log).memo,(*in_obs_log).memo);
  strcpy((*out_obs_log).comment,(*in_obs_log).comment);

  return;
}
void	StarCopyHistory(out_history,in_history)
HISTORY		*out_history,*in_history;
{
  unsigned int ii;

  (*out_history).size=(*in_history).size;
  strcpy((*out_history).step,(*in_history).step);
  (*out_history).time=(*in_history).time;
  for(ii=0;ii<(*in_history).size;ii++) {
    strcpy((*out_history).prm[ii],(*in_history).prm[ii]);
  }

  return;
}
void	StarCopySpurious(data_ch,out_spurious,in_spurious)
unsigned int	data_ch;
unsigned char	*out_spurious,*in_spurious;
{
  unsigned int sp_ch;
  unsigned int	ii;

  sp_ch=(data_ch-1)/(sizeof(unsigned char)*8)+1;
  for(ii=0;ii<sp_ch;ii++)
    out_spurious[ii]=in_spurious[ii];

  return;
}
void	StarCopyBaseline(data_ch,out_baseline,in_baseline)
unsigned int	data_ch;
unsigned char	*out_baseline,*in_baseline;
{
  unsigned int sp_ch;
  unsigned int	ii;

  sp_ch=(data_ch-1)/(sizeof(unsigned char)*8)+1;
  for(ii=0;ii<sp_ch;ii++)
    out_baseline[ii]=in_baseline[ii];

  return;
}
void	StarCopyProfdata(data_ch,out_data,in_data)
unsigned int	data_ch;
float	*out_data,*in_data;
{
  unsigned int	ii;

  for(ii=0;ii<data_ch;ii++)
    out_data[ii]=in_data[ii];

  return;
}

/* set the default head */
void StarSetDefaultProfile(profile)
PROFILE *profile;
/*
least setting for profile
2000/2/29 T.Handa
*/
{
  StarDefaultSyshead(&((*profile).sys_head));
  StarDefaultObsprm(&((*profile).obs_prm));
  StarDefaultObject(&((*profile).object));
  StarDefaultAntStatus(&((*profile).ant_status));
  StarDefaultRxStatus(&((*profile).rx_status));
  StarDefaultBeStatus(&((*profile).be_status));
  StarDefaultObsLog(&((*profile).obs_log));

  StarDefaultHistory(profile);

  return;
}

void StarDefaultSyshead(sys_head)
SYS_HEAD *sys_head;
/*
1999/4/5 T.Handa
*/
{
  (*sys_head).l_file=SCAN_INITIAL_L_HEAD; /* file length */
  (*sys_head).l_head=(*sys_head).l_file; /* header part length */
  (*sys_head).format_id=SPECTRUM;
  (*sys_head).version=SCAN_FORMAT_VERSION;
  (*sys_head).data_type=TA_DATA;
  (*sys_head).hist_step=0;
  (*sys_head).hist_line=0;
  (*sys_head).scan=0;
  (*sys_head).backup=0;
  (*sys_head).num_scan=0;

  return;
}

void StarDefaultObsprm(obsprm)
OBS_PRM *obsprm;
/*
1999/4/5 T.Handa
*/
{
  strcpy((*obsprm).project,"");
  strcpy((*obsprm).scd_file,"");
  (*obsprm).line_numb=(unsigned short)0;
  (*obsprm).sw_mode=ONOFF_SW;
  (*obsprm).mjd=0L;
  (*obsprm).start=0L;
  (*obsprm).stop=0L;
  (*obsprm).on_time=(unsigned short)0;
  (*obsprm).off_time=(unsigned short)0;
  strcpy((*obsprm).observer1,"");
  strcpy((*obsprm).observer2,"");
  strcpy((*obsprm).observer3,"");
  strcpy((*obsprm).observer4,"");

  return;
}

void StarDefaultObject(obj)
OBJECT *obj;
/*
1999/4/5 T.Handa
*/
{
  strcpy((*obj).name,"");
  (*obj).cood_sys=COORD_LB;
  (*obj).on_lb.x=0.0;
  (*obj).on_lb.y=0.0;
  (*obj).on_radec.x=0.0;
  (*obj).on_radec.y=0.0;
  strcpy((*obj).equinox,"B1950.0");
  (*obj).on_azel.x=0.0;
  (*obj).on_azel.y=0.0;
  (*obj).off_lb.x=0.0;
  (*obj).off_lb.y=0.0;
  (*obj).off_radec.x=0.0;
  (*obj).off_radec.y=0.0;
  (*obj).orig_xy.x=0.0;
  (*obj).orig_xy.y=0.0;
  (*obj).pa_xy=0.0;
  (*obj).on_xy.x=0.0;
  (*obj).on_xy.y=0.0;
  (*obj).azel_offset.x=0.0;
  (*obj).azel_offset.y=0.0;

  return;
}

void StarDefaultAntStatus(ant)
ANT_STATUS *ant;
/*
1999/4/5 T.Handa
*/
{
  strcpy((*ant).tele,"");
  (*ant).ap_eff=(float)0.0;
  (*ant).mb_eff=(float)0.0;
  (*ant).fss_eff=(float)0.0;
  (*ant).hpbw=(float)0.0;

  return;
}

void StarDefaultRxStatus(rx)
RX_STATUS *rx;
/*
1999/4/5 T.Handa
*/
{
  strcpy((*rx).name,"");
  (*rx).sb_mode=(char)NULL|SSB|USB;
  (*rx).v_lsr=0.0;
  (*rx).v_off=0.0;
  (*rx).rest_freq=0.0;
  (*rx).obs_freq=0.0;
  (*rx).lo_freq1=0.0;
  (*rx).if_freq1=0.0;
  (*rx).lo_freq2=0.0;
  (*rx).if_freq2=0.0;
  (*rx).lo_freq3=0.0;
  (*rx).if_freq3=0.0;
  (*rx).mltplx1=(unsigned short)0;
  (*rx).mltplx2=(unsigned short)0;
  (*rx).mltplx3=(unsigned short)0;
  (*rx).tsys=(float)0.0;
  (*rx).rms=(float)0.0;
  (*rx).calib=(float)0.0;
  (*rx).weight=(float)1.0;
  (*rx).scale=(float)1.0;

  return;
}

void StarDefaultBeStatus(be)
BE_STATUS *be;
/*
1999/4/5 T.Handa
*/
{
  strcpy((*be).name,"");
  (*be).be_ch=(unsigned short)1;
  (*be).basis_ch=(unsigned short)0;
  (*be).sign=(char)1;
  (*be).data_ch=(unsigned short)1;
  (*be).center_ch=0.0;
  (*be).center_freq=0.0;
  (*be).disp0=(*be).center_freq;
  (*be).disp1=1.0;
  (*be).disp2=0.0;
  (*be).disp3=0.0;
  (*be).disp4=0.0;
  (*be).freq_res=0.0;
  (*be).ch_width=(float)1.0;
  (*be).ch_orig=(float)1.0;

  return;
}

void StarDefaultObsLog(log)
OBS_LOG *log;
/*
1999/4/5 T.Handa
*/
{
  strcpy((*log).site_name,"");
  (*log).meteo.tmp=(float)0.0;
  (*log).meteo.water=(float)0.0;
  (*log).meteo.press=(float)0.0;
  (*log).tau0=(float)0.0;
  strcpy((*log).memo,"");
  strcpy((*log).comment,"");

  return;
}

void StarDefaultHistory(profile)
PROFILE *profile;
/*
2000/2/29 T.Handa
*/
{
  HISTORY hist;

  hist.size=0;
  strcpy(hist.step,"END");
  StarAddHistory(profile,&hist);

  return;
}
